// Oo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oOo.oO

#include <stdio.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/comm.h>
#include <nuon/intr.h>
#include "sprite2.h"

#define MAX_SPRITES 512
#define MAX_MPES 3

extern int spritec_start[],spritec_size[];
extern int sprited_start[],sprited_size[];
extern volatile int SpriteBase;

RASTER_INFO rinfo;
RASTER_INFO minfo[MAX_MPES];

SPRITE spriteBlock[2][MAX_SPRITES];
SPRITE spriteEmpty[MAX_SPRITES];
PriorityQueue freeSprites;
PriorityQueue usedSprites;
PriorityQueue displayList;

int startMPE,endMPE;
volatile long MPEMask;
long idleMask;
int sliceH;
int sprFlags;

int SPRCommHandler(int,int,int,int,int,int,int,int);

int GetTGASize(unsigned char *tga, long *h, long *w);
int ReadTGA(unsigned char *tga, long dmaflags, void *base,
        int xOutOffset, int yOutOffset, unsigned long *clut,long tcolRGB);

void inline _timeCheckpoint(long *usec, long *usecPrev, int print)
{
  if (print) {
    _TimeElapsed(NULL, usec);
    if (*usec < *usecPrev) *usec += 1000000;
    *usec -= *usecPrev;
    printf("  %d.%dms", (int) *usec / 1000, (int) (*usec % 1000) / 100);
  }
  _TimeElapsed(NULL, usecPrev);
}

void SPRFooLinkIt(void)
{
	int dummy1;
	dummy1 = SpriteBase;
}

void SPRFill(int initx, int inity, int wide, int high, int color)
{
    int x, y;
    int lastx, lasty;

    lastx = initx + wide;
    lasty = inity + high;

    for (x = initx; x < lastx; x += 8) {
		for (y = inity; y < lasty; y += 8) {
	    	_raw_plotpixel(rinfo.dmaflags, (void *)rinfo.dest, (8<<16)|x, (8<<16)|y, color);
		}
    }
}

int SPRInit(int startmpe, int endmpe, int sliceHeight, int flags)
{
    int i;

    startMPE = startmpe;
    endMPE   = endmpe;
    sliceH   = sliceHeight;
    sprFlags = flags;

    for(i = startMPE; i <= endMPE; i++) {
        MPEMask |= (1<<i);
        _MPELoad(i, (void *)0x20300000, spritec_start, (long) spritec_size);
        _MPELoad(i, (void *)0x20100000, sprited_start, (long) sprited_size);
        _MPERun (i, (void *)0x20300000);
    }
    idleMask = MPEMask;

    QInitMem(spriteBlock[0], MAX_SPRITES, sizeof(SPRITE));
    QInit(&freeSprites, (PriorityQueueMember *) spriteBlock[0]);
    QInit(&usedSprites, 0L);
    QInit(&displayList, 0L);
    for (i=0;i<MAX_SPRITES;i++) spriteBlock[0][i].parent = &freeSprites;

    memcpy(spriteEmpty, spriteBlock[0], sizeof(SPRITE) * MAX_SPRITES);

    _IntSetVector(kIntrCommRecv,SPRCommHandler);
    return 0;
}

int SPRReset()
{
  int i;

  QInitMem(spriteBlock[0], MAX_SPRITES, sizeof(SPRITE));
  QInit(&freeSprites, (PriorityQueueMember *) spriteBlock[0]);
  QInit(&usedSprites, 0L);
  QInit(&displayList, 0L);
  for (i = 0; i < MAX_SPRITES; i++) spriteBlock[0][i].parent = &freeSprites;

  return 0;
}

int SPRSetDestScreen(void *scr, long dmaflags, int minx, int miny, int maxx, int maxy, int bgColor)
{
    int i,h,sh,y;
    rinfo.dmaflags=dmaflags;
    rinfo.dest=scr;
    rinfo.clip_max_x=maxx;
    rinfo.clip_max_y=maxy;
    rinfo.clip_min_x=minx;
    rinfo.clip_min_y=miny;
    rinfo.bgColor=bgColor;

    h=maxy-miny+1;
    sh=h/(endMPE-startMPE+1);
    y=miny;
    for(i=startMPE;i<=endMPE;i++)
    {
        minfo[i].dmaflags=dmaflags;
        minfo[i].dest=scr;
        minfo[i].clip_min_x=minx;
        minfo[i].clip_max_x=maxx;
        minfo[i].clip_min_y=y;
        minfo[i].clip_max_y=y+sh-1;

        y+=sh;
    }
    return 0;
}

/*
   Sets up a TGA image to be usable as sprite source. The TGA is converted to either
   a 16 bpp or 32 bpp internal pixmap. The pixmap is setup in SDRAM. SDRAM memory is
   allocated via _MemAlloc
   
   tga:  TGA image
   mode: 0 = 16 bbp, 1 = 32 bpp 

*/

int SPRInstallTGAImage(char *tga, int mode, SPR_IMAGE_INFO *info,int transColorRGB)
{
	char *scr;
	long width,height,dmafl,type;
	
	GetTGASize(tga,&width,&height);
	
	type  = mode ? 4 : 2;  /* depth in bytes and DMA pixel type are conveniently the same !*/
	
	scr=_MemAlloc(width*height*type,512,kMemSDRAM);
	dmafl= (((width/8)<<16) | (type<<4) | DMA_PIXEL_WRITE | DMA_CLUSTER_BIT);
	
	ReadTGA(tga,dmafl,scr,0,0,0,transColorRGB);
	
	info->img = scr;
	info->dmaflags = dmafl | (1<<13);
	info->width = width;
	info->height = height;
	
	return 0;
}

/*
   Initializes a SPR_IMAGE_INFO structure with user defined values

*/

int SPRSetSourceImage(char *scr, int dmafl, int width, int height, SPR_IMAGE_INFO *info)
{
	info->img = scr;
	info->dmaflags = dmafl | (1<<13);
	info->width = width;
	info->height = height;
	
	return 0;
}


/*
    Creates a sprite. A sprite is defined by giving a rectangular area
    in a sprite source image (installed via SPRInstallTGAImage).
    
    The created sprite is not yet in the the display list and therefore invisible!
    
    Returns a sprite handle
*/

SPRITE *SPRCreateSprite(SPR_IMAGE_INFO *info,int x, int y, int w, int h)
{
  SPRITE *newspr;

  newspr = (SPRITE *)QGetHead(&freeSprites);
  if (newspr) {
    //        memset(newspr,0,sizeof(SPRITE));
    newspr->src_x      = x<<16;
    newspr->src_y      = y<<16;
    newspr->w          = w;
    newspr->h          = h;
    newspr->source     = info->img;
    newspr->dmaflags   = info->dmaflags;
    newspr->tint       = 0x10808000;
    newspr->transColor = 0x00808000;
    newspr->q.priority = 0;
    newspr->parent     = &usedSprites;

    QAdd(&usedSprites, &newspr->q, -1);

    return newspr;
  }

  return 0;
}

/*
    Clones an existing sprite.
*/

SPRITE *SPRCloneSprite(SPRITE *s)
{
    SPRITE *newspr;

    newspr = (SPRITE *)QGetHead(&freeSprites);
    if (newspr) {
        memcpy(newspr,s,sizeof(SPRITE));

        QAdd(&usedSprites, &newspr->q, sprFlags & SPR_FLAG_ADD_REVERSE);

        return newspr;
    }

    return 0;
}

/*
    Add a sprite created with SPRCreateSprite to the display list.
*/

int SPRAddSprite(SPRITE * sprite, int x, int y, int angle, int xscale, int yscale, int type,
		 int tc, int depth)
{
    if (sprite->parent != &usedSprites) return -1;

    QRemove(&usedSprites, &sprite->q);

    sprite->x          = x;
    sprite->y          = y;
    sprite->xscale     = xscale;
    sprite->yscale     = yscale;
    sprite->angle      = angle;
    sprite->trans      = type;
    sprite->transColor = tc;
    sprite->q.priority = depth;
    QAdd(&displayList, &sprite->q, sprFlags & SPR_FLAG_ADD_REVERSE);
    sprite->parent     = &displayList;

    return 0;
}

//
// Combination of SPRCreateSprite and SPRAddSprite
//
SPRITE * SPRCreateAndAddSprite(SPR_IMAGE_INFO *info, int srcX, int srcY, int srcW, int srcH,
			       int x, int y, int angle, int xscale, int yscale, int type,
			       int tc, int depth)
{
  SPRITE *newspr;

  newspr = (SPRITE *)QGetHead(&freeSprites);
  if (newspr) {
    newspr->src_x      = srcX<<16;
    newspr->src_y      = srcY<<16;
    newspr->w          = srcW;
    newspr->h          = srcH;
    newspr->source     = info->img;
    newspr->dmaflags   = info->dmaflags;

    newspr->x          = x;
    newspr->y          = y;
    newspr->xscale     = xscale;
    newspr->yscale     = yscale;
    newspr->angle      = angle;
    newspr->trans      = type;
    newspr->transColor = tc;
    newspr->q.priority = depth;
    QAdd(&displayList, &newspr->q, sprFlags & SPR_FLAG_ADD_REVERSE);
    newspr->parent     = &displayList;

    return (newspr);
    
  }

  return(0);
}

/*
    Modifies a sprite parameters
*/


int SPRModifySprite(SPRITE * sprite, int x, int y, int angle, int xscale, int yscale,
		    int type, int tc, int depth)
{
  if (sprite->q.priority != depth) QRemove(sprite->parent, &sprite->q);

  sprite->x=x;
  sprite->y=y;
  sprite->xscale=xscale;
  sprite->yscale=yscale;
  sprite->angle=angle;
  sprite->trans=type;
  sprite->transColor=tc;

  if (sprite->q.priority != depth) {
    sprite->q.priority=depth;
    if (sprite->parent == &displayList)
      QAdd(sprite->parent, &sprite->q, sprFlags & SPR_FLAG_ADD_REVERSE);
    else
      QAdd(sprite->parent, &sprite->q, -1);
  }
  return 0;
}

/*
    Removes a sprite from display list
*/

int SPRRemoveSprite(SPRITE *sprite)
{
    if (sprite->parent != &displayList) return -1;

    QRemove(&displayList, &sprite->q);
    QAdd(&usedSprites, &sprite->q, -1);
    sprite->parent=&usedSprites;

    return 0;
}

/*
    Deletes the sprite the sprite definition
*/

int SPRDeleteSprite(SPRITE *sprite)
{
    if (sprite->parent == &freeSprites) return 0;

    QRemove(&usedSprites, &sprite->q);
    QAdd(&freeSprites, &sprite->q, -1);
    sprite->parent=&freeSprites;
    
    return 0;
}

//
// Combination of previous two routines
//

int SPRRemoveAndDeleteSprite(SPRITE *sprite)
{
    if (sprite->parent != &displayList) return -1;

    QRemove(&displayList, &sprite->q);
    QAdd(&freeSprites, &sprite->q, -1);
    sprite->parent = &freeSprites;

    return 0;
}

/*
    Quick simple parameter modifiers
*/


void SPRSetSpriteSource(SPRITE * sprite, int x, int y, int w, int h)
{
    sprite->src_x=x<<16;
    sprite->src_y=y<<16;
    sprite->w=w;
    sprite->h=h;
}

void SPRSetSpriteXY(SPRITE * sprite, int x, int y)
{
    sprite->x=x;
    sprite->y=y;
}

void SPRSetSpriteScale(SPRITE * sprite, int xscale, int yscale)
{
    sprite->xscale=xscale;
    sprite->yscale=yscale;
}

void SPRSetSpriteRotation(SPRITE * sprite, int angle)
{
  sprite->angle=angle;
}

void SPRSetSpriteType(SPRITE * sprite, int type)
{
    sprite->trans=type;
}

void SPRSetSpriteTColor(SPRITE * sprite, int tc)
{
    sprite->transColor=tc;
}

void SPRSetSpriteDepth(SPRITE * sprite, int depth)
{
    QRemove(sprite->parent, &sprite->q);
    sprite->q.priority=depth;
    if (sprite->parent == &displayList)
      QAdd(sprite->parent, &sprite->q, sprFlags & SPR_FLAG_ADD_REVERSE);
    else
      QAdd(sprite->parent, &sprite->q, -1);
}


/*
    Main drawing function
    
    Walks down the display list and feed the available MPEs to render the sprites
    
    Commbus command for the renderes:
    
    p[0]:  Pointer to RASTER_INFO structure (destination info and clip region)
    p[1]:  Pointer to SPRITE
//    p[2]:  Bit 16-31: Slice height, Bit 0: One-Shot (draw only one sprite)
    p[2]:  Bit 24-31: Slice height, Bit 16-23: Logical MPE number, Bit 8-15: Total MPE number, 
           Bit 0: One-Shot (draw only one sprite)
//    p[3]:  Bit 16-31: Logical MPE number, Bit 0-15: Total MPE number 
    p[3]:  Clear color
*/



void SPRDraw(int clrScreen, int wait)
{
#undef NOTIME
#define NOTIME

    int m,sr,i;
#ifndef NOTIME
    long usec, usecPrev;
#endif

    if (!displayList.list) return;
    
    // Wait until the renderers are finished
    while (MPEMask != idleMask);
      
    if (sprFlags & SPR_FLAG_DOUBLE_BUFFER) {
#ifndef NOTIME
      printf ("**");
      _timeCheckpoint(&usec, &usecPrev, 0);
#endif

      memcpy(spriteBlock[1], spriteBlock[0], sizeof(SPRITE) * MAX_SPRITES);
      //    printf("\n START: %08X %08X ", spriteBlock[0], spriteBlock[1]);
      for (i = 0; i < MAX_SPRITES; i++) {
	if (spriteBlock[1][i].q.next)
	  spriteBlock[1][i].q.next =
	    (PriorityQueueMember*) ((long) spriteBlock[1][i].q.next + sizeof(SPRITE) * MAX_SPRITES);
	//      printf("\n %08X %08X ", spriteBlock[0][i].q.next.index, spriteBlock[1][i].q.next.index);
      }

#ifndef NOTIME
      _timeCheckpoint(&usec, &usecPrev, 1);
#endif
    }

    _DCacheSync();

    if (clrScreen) {
    /*
        SPRFill(rinfo.clip_min_x,rinfo.clip_min_y,
            rinfo.clip_max_x-rinfo.clip_min_x+1,
            rinfo.clip_max_y-rinfo.clip_min_y+1,rinfo.bgColor);
    */
      if (sprFlags & SPR_FLAG_CLEAR_FAST)
	clrScreen = 1<<2;
      else {
        SPRFill(rinfo.clip_min_x, rinfo.clip_min_y, rinfo.clip_max_x-rinfo.clip_min_x+1,
                rinfo.clip_max_y-rinfo.clip_min_y+1, rinfo.bgColor);
        clrScreen = 0;
      }

    }

    DISABLE_INTERRUPTS(sr);
    MPEMask = 0x0;
    RESTORE_INTERRUPTS(sr);
    for (m = startMPE, i = 0; m <= endMPE; m++, i++) {
      //        _CommSendDirect((long)&rinfo, (long)(displayList.list),
      //			sliceH<<16, (i<<16)+(endMPE-startMPE+1),m,0);
      
      if (sprFlags & SPR_FLAG_DOUBLE_BUFFER) {
        _CommSendDirect((long) &rinfo, ((long) displayList.list) + sizeof(SPRITE) * MAX_SPRITES,
                        (sliceH<<24) | (i<<16) | ((endMPE-startMPE+1)<<8) | clrScreen,
                        rinfo.bgColor, m, 0);
      } else {
        _CommSendDirect((long) &rinfo, (long) displayList.list,
                        (sliceH<<24) | (i<<16) | ((endMPE-startMPE+1)<<8) | clrScreen,
                        rinfo.bgColor, m, 0);
      }
    }

    /* Wait for them to finish */
    if (wait) while (MPEMask != idleMask);
}

int SPRWait(void)
{
  int cnt = 0;

  while (MPEMask != idleMask) { cnt++; };

  return(cnt);
}


/*
   Low-Level blitter call
*/

SPRITE srcinfo;
RASTER_INFO destinfo;

int SPRBlitter(char *src, long src_dmaflags, int src_x, int src_y, int src_w, int src_h,
                char *dest, long dest_dmaflags, int dest_x, int dest_y,
                int scalex, int scaley, int angle,
                int clip_min_x,int clip_min_y,int clip_max_x,int clip_max_y,
                int type, int transColor)
{
    int sr,m,i;
	
    srcinfo.q.next=0;
    srcinfo.src_x = src_x<<16;
    srcinfo.src_y = src_y<<16;
    srcinfo.x = dest_x;
    srcinfo.y = dest_y;
    srcinfo.xscale = scalex;
    srcinfo.yscale = scaley;
    srcinfo.w = src_w;
    srcinfo.h = src_h;
    srcinfo.source=src;
    srcinfo.dmaflags=src_dmaflags;
    srcinfo.trans=type;
    srcinfo.tint=0x10808000;
    srcinfo.transColor=transColor;

    destinfo.dest=dest;
    destinfo.dmaflags=dest_dmaflags;
    destinfo.clip_min_x=clip_min_x;
    destinfo.clip_min_y=clip_min_y;
    destinfo.clip_max_x=clip_max_x;
    destinfo.clip_max_y=clip_max_y;

    _DCacheSync();
    DISABLE_INTERRUPTS(sr);
    MPEMask = 0x0;
    RESTORE_INTERRUPTS(sr);
    for (m=startMPE,i=0;m<=endMPE;m++,i++)
      //        _CommSendDirect((long)&destinfo,(long)&(srcinfo),(sliceH<<16)+1,(i<<16)+(endMPE-startMPE+1),m,0);
      _CommSendDirect((long)&destinfo, (long)&(srcinfo),
                      (sliceH<<24) | (i<<16) | ((endMPE-startMPE+1)<<8) | (1<<0),
                      destinfo.bgColor, m,0);
	while (MPEMask != idleMask);
	return 0;
}	
