;
; Input parameters
;
        sine__x	= r0
;
; Results
;
	cosine	= r0
        sine	= r1
;
; Implementation ====================
;
; Working register declarations
;
	quadrant	= r0
	sine__one		= r3
	q1		= r7
	q2		= r4
	sine__temp		= r5
	sine__y1		= r1
	sine__y2		= r6
	sine__x1		= r2
	sine__x2		= r3
;
; Value definitions
;
	quadShift	= 14		; shift to right-adjust quadrant "field"
	quad2or4Shift	= 0		; bits to right of bit we want
	quad3or4Mask	= 2		; mask
;
; Coefficients for cubic polynomial evaluation
;
	A0 = fix(1.57079101108022,30)
	A1 = fix(-0.645892849558497,30)
	A2 = fix(0.0794343446081996,30)
	A3 = fix(-0.00433309527758793,30)


sincos:
{	mv_s	#A3,sine__y1			; load the first coefficient
	lsl	#18,sine__x,sine__x1		; normalize 1/4-rotation to 0.32
}
{	jmp	eq,rightAngle,nop	; jump if exact right angle
	mv_s	#fix(1,30),sine__one		; load constant 1 in 2.30 format
	lsr	#2,sine__x1			; adjust 1/4-rotation to 2.30
}
{	mv_s	sine__x1,q1			; save q1 = adjusted quarter-rotation
	mul	sine__x1,sine__x1,>>#30,sine__x1		; sine__x1 = q1 ^ 2
	sub	sine__x1,sine__one,sine__x2		; calculate "opposite" 1/4-rotation
}
{	mul	sine__x2,sine__x2,>>#30,sine__x2		; sine__x2 = q2 ^ 2
	mv_s	sine__x2,q2			; save q2 = adjusted quarter-rotation
}
;
; Evaluate the polynomial using Horner's method.  Three things are going
; on in parallel:
;
;  1. Horner's method is used to calculate a 1/4 rotation result.
;  2. Horner's method is used to calculate opposite 1/4 rotation result.
;  3. A jump table is being processed to handle the two results based
;     on quadrant.
;
{	mul	sine__x1,sine__y1,>>#30,sine__y1
	mv_s	sine__y1,sine__y2
	lsr	#quadShift - 2,sine__x,quadrant ; save the "quadrant" field of sine__x
}
{	mul	sine__x2,sine__y2,>>#30,sine__y2
	mv_s	#A2,sine__temp
	and	#3 << 2,quadrant	; clear unwanted bits from quadrant
}
	add	sine__temp,sine__y1
{	mv_s	#jmpTable,sine__temp		; get address of jump table
	mul	sine__x1,sine__y1,>>#30,sine__y1
	add	sine__temp,sine__y2
}
{	mul	sine__x2,sine__y2,>>#30,sine__y2
	mv_s	#A1,sine__temp
	add	sine__temp,quadrant		; index into jump table
}
	add	sine__temp,sine__y1
{	ld_s	(quadrant),quadrant	; get address from jump table
	mul	sine__x1,sine__y1,>>#30,sine__y1
	add	sine__temp,sine__y2
}
	mul	sine__x2,sine__y2,>>#30,sine__y2
	mv_s	#A0,sine__temp
{	jmp	(quadrant)
	add	sine__temp,sine__y1
}
{	mul	q1,sine__y1,>>#30,sine__y1		; multiply sine__y1 polynomial by q1
	add	sine__temp,sine__y2
}
	mul	q2,sine__y2,>>#30,sine__y2		; multiply sine__y2 polynomial by q2
;
; The following code segments are reached via the jump table, with
; a segment for each quadrant of a rotation.
;
quad1:
	rts
	mv_s	sine__y1,sine			; sine = sine__y1
	mv_s	sine__y2,cosine		; cosine = sine__y2
quad2:
	rts
	sub	sine__y1,#0,cosine		; cosine = -sine__y1
	mv_s	sine__y2,sine			; sine = sine__y2
quad3:
	rts
	sub	sine__y1,#0,sine		; sine = -sine__y1
	sub	sine__y2,#0,cosine		; cosine = -sine__y2
quad4:
	rts
	mv_s	sine__y1,cosine		; cosine = sine__y1
	sub	sine__y2,#0,sine		; sine = -sine__y2
;
; Special case precise result for exact right angle.
;
rightAngle:
{	mv_s	#jmpTable,sine__temp		; get address of jump table
	lsr	#quadShift - 2,sine__x,quadrant ; save the "quadrant" field of sine__x
}
	and	#3 << 2,quadrant	; clear unwanted bits from quadrant
{	mv_s	#fix(1,30),sine__y2		; sine__y2 = 1
	add	sine__temp,quadrant		; calculate jump table address
}
	ld_s	(quadrant),quadrant	; load address from jump table
	nop
	jmp	(quadrant)		; jump through table
	sub	sine__y1,sine__y1			; sine__y1 = 0
	nop
