#include <stdio.h>
#include <stdlib.h>
#include "sprite.h"
#include <nuon/termemu.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>

#define SCREEN_WIDTH 360	
#define SCREEN_HEIGHT 240
#define DMAFLAGS  (DMA_PIXEL_WRITE | ((SCREEN_WIDTH/8)<<16) | (4<<4) | (1<<11))                                          

#define NUM_SPRITES 64
#define NUM_BUFFERS 3

extern char SpriteImage[];

void *FrameBuffer[NUM_BUFFERS];
static int fps;
int drawBuffer = 0;

/*
 * wait for video
 * returns the number of elapsed fields
 * since the last sync
 */

int
VideoSync(void)
{
    static int frames;
    static int lastfield;
    static int lastsecond;
    int ret;

    int curfield;

    /* we want to view the buffer we just finished
       drawing */

    /* set the video to point at the drawn buffer */
    _VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[drawBuffer]);

    /* find out how many fields have been drawn */
    curfield = _VidSync(-1);

    /* remember that we've drawn a frame */
    frames++;

    /* have more than 60 fields (1 second) passed? */
    if (curfield - lastsecond >= 60) {
        fps = frames;
        frames = 0;
        lastsecond = curfield;
    }

    /* draw to a new buffer */
    drawBuffer++;
    if (drawBuffer >= NUM_BUFFERS)
        drawBuffer = 0;

#if NUM_BUFFERS == 2
    /* wait until a new field starts (so it's no longer
       showing the field we're going to draw on) */
    curfield = _VidSync(0);
#endif

    ret = curfield - lastfield;
    while (ret == 0) {
        curfield = _VidSync(-1);
        ret = curfield - lastfield;
    }
    lastfield = curfield;
    return ret;
}

int main()
{
    SPR_IMAGE_INFO img;
    SPRITE *sprite[NUM_SPRITES];
    int x[NUM_SPRITES];
    int y[NUM_SPRITES];
    int ix[NUM_SPRITES];
    int iy[NUM_SPRITES];
    int angle[NUM_SPRITES];
    int ia[NUM_SPRITES];
    int i;

	VidDisplay display;
	VidChannel mainchannel;
    char temp[256];

    /* Setup video */

    for(i=0;i<NUM_BUFFERS;i++)
	    FrameBuffer[i]=_MemAlloc(SCREEN_WIDTH*SCREEN_HEIGHT*4,512,kMemSDRAM);

 	display.dispwidth = -1;
	display.dispheight = -1;
	display.bordcolor=0x20100000;
	display.progressive=0;
	for(i=0;i<6;i++)
		display.reserved[i]=0;
	
	mainchannel.dmaflags=DMAFLAGS;
	mainchannel.base=FrameBuffer[0];
	mainchannel.dest_xoff = -1;
	mainchannel.dest_yoff = -1;
	mainchannel.dest_width = 720;
	mainchannel.dest_height = 480;
	mainchannel.src_xoff = 0;
	mainchannel.src_yoff = 0;
	mainchannel.src_width = SCREEN_WIDTH;
	mainchannel.src_height = SCREEN_HEIGHT;
	mainchannel.clut_select=0;
	mainchannel.alpha=0;
	mainchannel.vfilter=VID_VFILTER_2TAP;
	mainchannel.hfilter=VID_HFILTER_4TAP;
	for(i=0;i<5;i++)
		mainchannel.reserved[i]=0;
   
    InitTerminalX(0,(int)FrameBuffer[0],SCREEN_WIDTH, SCREEN_HEIGHT, DMAFLAGS, kBlack);                                                              
    _VidConfig ( &display, &mainchannel, 0L, 0L); 
		
    /* Use MPEs 0,1 and 2 for rendering, 16 pixel slices */
    SPRInit(0,2,16);

    /* Set to the second framebuffer */
    SPRSetDestScreen(FrameBuffer[1], DMAFLAGS,  0,  0,  SCREEN_WIDTH-1,  SCREEN_HEIGHT-1, kBlack);
    
    /* Install the TGA image */
//    SPRInstallTGAImage(SpriteImage, 1, &img, 0xf0f0ee00);
    SPRInstallTGAImage(SpriteImage, 1, &img, 0x02020200);

    for(i=0;i<NUM_SPRITES;i++)
    {
        /* Create a few sprites */
        
        sprite[i]=SPRCreateSprite(&img,40*(rand()%3),0,32,32);
        
    
        x[i] = (rand() % 280) + 40;
        y[i] = (rand() % 160) + 40;
#if 1
        ix[i] = (rand() % 4) - 2;
        iy[i] = (rand() % 4) - 2;
#else
        ix[i] = 1;
        iy[i] = 1;
#endif
        angle[i]=0;
        ia[i] = (rand() % 0x100) + 0x100;

        /* Add sprites to display list*/    
        
        SPRAddSprite(sprite[i], x[i], y[i], 0, 0x8000, 0x8000, 0x40000002, kBlack, 1);
    }

    for(;;)
    {
      
        VideoSync();

        /* Process display list */   
	    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS,  0,  0,  SCREEN_WIDTH-1,  SCREEN_HEIGHT-1, kBlack);
        SPRDraw(1,0);

        /* print FPS */   
#if 1
        SetScreenBase((int)FrameBuffer[drawBuffer]);
        sprintf(temp,"%d fps",fps);
        PrintStatus(temp,kWhite,kBlack);
#endif        

#if 1
     /* Mess around with some sprite parameters */   
        
        for(i=0;i<NUM_SPRITES;i++)
        {

            x[i]+=ix[i];
            if (x[i]>=320 || x[i]<=40) {
                ix[i]=-ix[i];
                ia[i]=-ia[i];
            }
            y[i]+=iy[i];
            if (y[i]>=200 || y[i]<=40) {
                iy[i]=-iy[i];
                ia[i]=-ia[i];
            }
            angle[i]+=ia[i];
        }
        SPRWait();
        for(i=0;i<NUM_SPRITES;i++)
        {
            SPRSetSpriteXY(sprite[i],x[i],y[i]);
            SPRSetSpriteRotation(sprite[i],angle[i]);
        }
#endif        
        
    }

}
